"""Zeus is a framework for deep learning energy measurement and optimization.

- [`device`][zeus.device]: Abstraction layer over compute devices
- [`monitor`][zeus.monitor]: Programmatic power and energy measurement tools
- [`optimizer`][zeus.optimizer]: A collection of optimizers for time and energy
- [`utils`][zeus.utils]: Utility functions and classes
- [`callback`][zeus.callback]: Callback definition
- [`exception`][zeus.exception]: Base exception class definition
- [`metric`][zeus.metric]: Tools for defining and tracking power and energy-related metrics
- [`show_env`][zeus.show_env]: Command line tool for install verification and device detection
- [`_legacy`][zeus._legacy.policy]: Legacy code mostly to keep our papers reproducible
"""

__version__ = "0.12.1.post1"
